//--------------------------------------------------------------------
// YC48YC48̐FԁiFjϊ
//
// BT.709BT.601
// yy = Y + ((-15640*U - 27905*V + 65536) >> 17);
// uu = U + ((  2456*U + 15038*V + 65536) >> 17);
// vv = V + ((  9931*U +  3352*V + 65536) >> 17);
//
// BT.601BT.709
// yy = Y + ((13441*U + 25705*V + 65536) >> 17);
// uu = U + ((-1331*U - 14514*V + 65536) >> 17);
// vv = V + ((-9585*U -  2196*V + 65536) >> 17);
//
//--------------------------------------------------------------------
#include <stdint.h>
#include <windows.h>
#include <mmintrin.h>
#include <xmmintrin.h>
#include <emmintrin.h>
#include "color.h"

//---------------------------------------------------------------------
// YC48Fϊ̋ʕ
//
// convert_mode
// 0: BT.709BT.601
// 1: BT.601BT.709
//---------------------------------------------------------------------
void __attribute__((force_align_arg_pointer)) my_convert( PIXEL_YC *dst, const PIXEL_YC *src, const int w, const int src_size, const int convert_mode )
{
    // BT.709BT.601
    // yy = Y + ((13441*U + 25705*V + 65536) >> 17)
    // uu = U + ((-1331*U - 14514*V + 65536) >> 17)
    // vv = V + ((-9585*U -  2196*V + 65536) >> 17)
    //                             
    //               coe0         coeA
    //                 coe1
    //                 coe2
    //
    // BT.601BT.709
    // yy = Y + ((-15640*U - 27905*V + 65536) >> 17)
    // uu = U + ((  2456*U + 15038*V + 65536) >> 17)
    // vv = V + ((  9931*U +  3352*V + 65536) >> 17)
    //                             
    //               coe0         coeA
    //                 coe1
    //                 coe2
    const __m128i coe0 = (convert_mode == 0)
                        ? _mm_setr_epi16(  13441,  25705, -1331, -14514, -9585, -2196,   13441,  25705 )
                        : _mm_setr_epi16( -15640, -27905,  2456,  15038,  9931,  3352,  -15640, -27905 );
    const __m128i coe1 = (convert_mode == 0)
                        ? _mm_setr_epi16( -1331, -14514, -9585, -2196,  13441,  25705, -1331, -14514 )
                        : _mm_setr_epi16(  2456,  15038,  9931,  3352, -15640, -27905,  2456,  15038 );
    const __m128i coe2 = (convert_mode == 0)
                        ? _mm_setr_epi16( -9585, -2196,  13441,  25705, -1331, -14514, -9585, -2196 )
                        : _mm_setr_epi16(  9931,  3352, -15640, -27905,  2456,  15038,  9931,  3352 );
    const __m128i coeA = _mm_set1_epi32( 65536 );

    int x = 0;

    // o͑̃|C^16byteACgɑB
    // 蔲̂߁ApaddingŒł2byteɑĂ  xȏ̉KvB
    intptr_t padding = (-(intptr_t)dst) & 0xF;
    if( w > 64 && !(padding & 0x1) )
    {
        padding = ((padding >> 1) * 3) & 0x7; // ̂ɕKvȃsNZ

        _mm_prefetch( (BYTE *)src + src_size, _MM_HINT_NTA ); // ̃CvtFb`

        // 1sNZo͂đ
        for( ; x < (int)padding; x++ )
        {
            __m128i t0 = _mm_setzero_si128();
            __m128i t1 = _mm_setzero_si128();
            t0 = _mm_insert_epi16( t0, src->y,  0 );
            t0 = _mm_insert_epi16( t0, src->cb, 1 );
            t0 = _mm_insert_epi16( t0, src->cr, 2 );
            t1 = _mm_insert_epi16( t1, src->cb, 0 );
            t1 = _mm_insert_epi16( t1, src->cr, 1 );
            t1 = _mm_shuffle_epi32( t1, _MM_SHUFFLE( 0, 0, 0, 0 ) );
            t1 = _mm_madd_epi16( t1, coe0 );
            t1 = _mm_add_epi32( t1, coeA );
            t1 = _mm_srai_epi32( t1, 17 );
            t1 = _mm_packs_epi32( t1, _mm_setzero_si128() );
            t0 = _mm_add_epi16( t0, t1 );
            dst->y  = _mm_extract_epi16( t0, 0 );
            dst->cb = _mm_extract_epi16( t0, 1 );
            dst->cr = _mm_extract_epi16( t0, 2 );
            src++;
            dst++;
        }
        // o͑̃ACgĂ
        // 8sNZ
        for( ; x < w - 7; x += 8 )
        {
            _mm_prefetch( (BYTE *)src + src_size, _MM_HINT_NTA ); // ̃CvtFb`

            const __m128i in0 = _mm_loadu_si128( (__m128i *)src     );
            const __m128i in1 = _mm_loadu_si128( (__m128i *)src + 1 );
            const __m128i in2 = _mm_loadu_si128( (__m128i *)src + 2 );

            // яiʂj
            // in0: Y0 U0 V0 Y1 U1 V1 Y2 U2
            // in1: V2 Y3 U3 V3 Y4 U4 V4 Y5
            // in2: U5 V5 Y6 U6 V6 Y7 U7 V7
            {
                // t0: U0 V0 U0 V0 U0 V0 U1 V1
                // t1: U1 V1 U1 V1 U2 V2 U2 V2
                __m128i t0 = _mm_shufflelo_epi16( in0, _MM_SHUFFLE( 2, 1, 2, 1 ) );
                t0 = _mm_shuffle_epi32( t0, _MM_SHUFFLE( 2, 0, 1, 0 ) );
                __m128i t1 = _mm_insert_epi16( in0, _mm_extract_epi16( in0, 7 ), 0 );
                t1 = _mm_insert_epi16( t1, _mm_extract_epi16( in1, 0 ), 1 );
                t1 = _mm_shuffle_epi32( t1, _MM_SHUFFLE( 0, 0, 2, 2 ) );

                t0 = _mm_madd_epi16( t0, coe0 );
                t1 = _mm_madd_epi16( t1, coe1 );
                t0 = _mm_add_epi32( t0, coeA );
                t1 = _mm_add_epi32( t1, coeA );
                t0 = _mm_srai_epi32( t0, 17 );
                t1 = _mm_srai_epi32( t1, 17 );

                __m128i out0 = _mm_add_epi16( in0, _mm_packs_epi32( t0, t1 ) );
                _mm_stream_si128( (__m128i *)dst, out0 );
            }

            {
                // t2: U2 V2 U3 V3 U3 V3 U3 V3
                // t3: U4 V4 U4 V4 U4 V4 U5 V5
                __m128i t2 = _mm_insert_epi16( in1, _mm_extract_epi16( in1, 0 ), 1 );
                t2 = _mm_insert_epi16( t2, _mm_extract_epi16( in0, 7 ), 0 );
                t2 = _mm_shuffle_epi32( t2, _MM_SHUFFLE( 1, 1, 1, 0 ) );
                __m128i t3 = _mm_unpacklo_epi32( _mm_srli_si128( in1, 10 ), in2 );
                t3 = _mm_shuffle_epi32( t3, _MM_SHUFFLE( 1, 0, 0, 0 ) );

                t2 = _mm_madd_epi16( t2, coe2 );
                t3 = _mm_madd_epi16( t3, coe0 );
                t2 = _mm_add_epi32( t2, coeA );
                t3 = _mm_add_epi32( t3, coeA );
                t2 = _mm_srai_epi32( t2, 17 );
                t3 = _mm_srai_epi32( t3, 17 );

                __m128i out1 = _mm_add_epi16( in1, _mm_packs_epi32( t2, t3 ) );
                _mm_stream_si128( (__m128i *)dst + 1, out1 );
            }

            {
                // t4: U5 V5 U5 V5 U6 V6 U6 V6
                // t5: U6 V6 U7 V7 U7 V7 U7 V7
                __m128i t4 = _mm_shuffle_epi32( in2, _MM_SHUFFLE( 1, 2, 0, 0 ) );
                t4 = _mm_shufflehi_epi16( t4, _MM_SHUFFLE( 0, 3, 0, 3 ) );
                __m128i t5 = _mm_insert_epi16( in2, _mm_extract_epi16( in2, 3 ), 0 );
                t5 = _mm_insert_epi16( t5, _mm_extract_epi16( in2, 4 ), 1 );
                t5 = _mm_shuffle_epi32( t5, _MM_SHUFFLE( 3, 3, 3, 0 ) );

                t4 = _mm_madd_epi16( t4, coe1 );
                t5 = _mm_madd_epi16( t5, coe2 );
                t4 = _mm_add_epi32( t4, coeA );
                t5 = _mm_add_epi32( t5, coeA );
                t4 = _mm_srai_epi32( t4, 17 );
                t5 = _mm_srai_epi32( t5, 17 );

                __m128i out2 = _mm_add_epi16( in2, _mm_packs_epi32( t4, t5 ) );
                _mm_stream_si128( (__m128i *)dst + 2, out2 );
            }

            src += 8;
            dst += 8;
        }

    }
    else
    {
        // ACgĂȂ
        // 8sNZ
        for( ; x < w - 7; x += 8 )
        {
            _mm_prefetch( (BYTE *)src + src_size, _MM_HINT_NTA ); // ̃CvtFb`

            const __m128i in0 = _mm_loadu_si128( (__m128i *)src     );
            const __m128i in1 = _mm_loadu_si128( (__m128i *)src + 1 );
            const __m128i in2 = _mm_loadu_si128( (__m128i *)src + 2 );

            // яiʂj
            // in0: Y0 U0 V0 Y1 U1 V1 Y2 U2
            // in1: V2 Y3 U3 V3 Y4 U4 V4 Y5
            // in2: U5 V5 Y6 U6 V6 Y7 U7 V7
            {
                // t0: U0 V0 U0 V0 U0 V0 U1 V1
                // t1: U1 V1 U1 V1 U2 V2 U2 V2
                __m128i t0 = _mm_shufflelo_epi16( in0, _MM_SHUFFLE( 2, 1, 2, 1 ) );
                t0 = _mm_shuffle_epi32( t0, _MM_SHUFFLE( 2, 0, 1, 0 ) );
                __m128i t1 = _mm_insert_epi16( in0, _mm_extract_epi16( in0, 7 ), 0 );
                t1 = _mm_insert_epi16( t1, _mm_extract_epi16( in1, 0 ), 1 );
                t1 = _mm_shuffle_epi32( t1, _MM_SHUFFLE( 0, 0, 2, 2 ) );

                t0 = _mm_madd_epi16( t0, coe0 );
                t1 = _mm_madd_epi16( t1, coe1 );
                t0 = _mm_add_epi32( t0, coeA );
                t1 = _mm_add_epi32( t1, coeA );
                t0 = _mm_srai_epi32( t0, 17 );
                t1 = _mm_srai_epi32( t1, 17 );

                __m128i out0 = _mm_add_epi16( in0, _mm_packs_epi32( t0, t1 ) );
                _mm_storeu_si128( (__m128i *)dst, out0 );
            }

            {
                // t2: U2 V2 U3 V3 U3 V3 U3 V3
                // t3: U4 V4 U4 V4 U4 V4 U5 V5
                __m128i t2 = _mm_insert_epi16( in1, _mm_extract_epi16( in1, 0 ), 1 );
                t2 = _mm_insert_epi16( t2, _mm_extract_epi16( in0, 7 ), 0 );
                t2 = _mm_shuffle_epi32( t2, _MM_SHUFFLE( 1, 1, 1, 0 ) );
                __m128i t3 = _mm_unpacklo_epi32( _mm_srli_si128( in1, 10 ), in2 );
                t3 = _mm_shuffle_epi32( t3, _MM_SHUFFLE( 1, 0, 0, 0 ) );

                t2 = _mm_madd_epi16( t2, coe2 );
                t3 = _mm_madd_epi16( t3, coe0 );
                t2 = _mm_add_epi32( t2, coeA );
                t3 = _mm_add_epi32( t3, coeA );
                t2 = _mm_srai_epi32( t2, 17 );
                t3 = _mm_srai_epi32( t3, 17 );

                __m128i out1 = _mm_add_epi16( in1, _mm_packs_epi32( t2, t3 ) );
                _mm_storeu_si128( (__m128i *)dst + 1, out1 );
            }

            {
                // t4: U5 V5 U5 V5 U6 V6 U6 V6
                // t5: U6 V6 U7 V7 U7 V7 U7 V7
                __m128i t4 = _mm_shuffle_epi32( in2, _MM_SHUFFLE( 1, 2, 0, 0 ) );
                t4 = _mm_shufflehi_epi16( t4, _MM_SHUFFLE( 0, 3, 0, 3 ) );
                __m128i t5 = _mm_insert_epi16( in2, _mm_extract_epi16( in2, 3 ), 0 );
                t5 = _mm_insert_epi16( t5, _mm_extract_epi16( in2, 4 ), 1 );
                t5 = _mm_shuffle_epi32( t5, _MM_SHUFFLE( 3, 3, 3, 0 ) );

                t4 = _mm_madd_epi16( t4, coe1 );
                t5 = _mm_madd_epi16( t5, coe2 );
                t4 = _mm_add_epi32( t4, coeA );
                t5 = _mm_add_epi32( t5, coeA );
                t4 = _mm_srai_epi32( t4, 17 );
                t5 = _mm_srai_epi32( t5, 17 );

                __m128i out2 = _mm_add_epi16( in2, _mm_packs_epi32( t4, t5 ) );
                _mm_storeu_si128( (__m128i *)dst + 2, out2 );
            }

            src += 8;
            dst += 8;
        }
    }

     // c1sNZ
     for( ; x < w; x++ )
     {
         __m128i t0 = _mm_setzero_si128();
         __m128i t1 = _mm_setzero_si128();
         t0 = _mm_insert_epi16( t0, src->y,  0 );
         t0 = _mm_insert_epi16( t0, src->cb, 1 );
         t0 = _mm_insert_epi16( t0, src->cr, 2 );
         t1 = _mm_insert_epi16( t1, src->cb, 0 );
         t1 = _mm_insert_epi16( t1, src->cr, 1 );
         t1 = _mm_shuffle_epi32( t1, _MM_SHUFFLE( 0, 0, 0, 0 ) );
         t1 = _mm_madd_epi16( t1, coe0 );
         t1 = _mm_add_epi32( t1, coeA );
         t1 = _mm_srai_epi32( t1, 17 );
         t1 = _mm_packs_epi32( t1, _mm_setzero_si128() );
         t0 = _mm_add_epi16( t0, t1 );
         dst->y  = _mm_extract_epi16( t0, 0 );
         dst->cb = _mm_extract_epi16( t0, 1 );
         dst->cr = _mm_extract_epi16( t0, 2 );
         src++;
         dst++;
     }
}

//---------------------------------------------------------------------
// [Xbh] YC48() -> YC48iFϊꍇj
//
// convert_mode
// 0: BT.601 BT.709
// 1: BT.709 BT.601
//---------------------------------------------------------------------
void __attribute__((force_align_arg_pointer)) thread_pixel2yc_yc48_convert( int thread_id, int thread_num, void *param1, void *param2 )
{
    // l[
    const COLOR_PROC_INFO *cpip  = (COLOR_PROC_INFO *)param1;
    const int convert_mode = *(int *)param2;

    // [JϐɃRs[
    const int h         = cpip->h;
    const int w         = cpip->w;
    const int line_size = cpip->line_size;

    // ̂ŁAXbh𔼕ɂė̏Ƀ\[X
    if( thread_id < 0 || thread_num <= 0 ) return; // 肦ȂlȂ̂
    if( thread_id & 0x1 ) return; // eîݎcj
    thread_id = thread_id >> 1;
    thread_num = (thread_num + 1 ) >> 1;

    // Xbhc̃TCY葽璲
    if( thread_num > h ) thread_num = h;
    if( thread_id >= thread_num ) return;

    // Xbh̉摜郉CvZ
    const int y_start = (h *  thread_id     ) / thread_num;
    const int y_end   = (h * (thread_id + 1)) / thread_num;

    // ϐvZ
    const int pixel_line_size = (w * sizeof(PIXEL_YC) + 3) & (~3L); // 4̔{ɐ؂グ
    PIXEL_YC *ycp = cpip->ycp;
    const unsigned char *pixelp  = (unsigned char *)cpip->pixelp;


    for( int y = y_start; y < y_end; y++ )
    {
        const PIXEL_YC *pixelp2 = (PIXEL_YC *)( (unsigned char *)pixelp + y * pixel_line_size );
        PIXEL_YC *ycp2 = (PIXEL_YC *)( (BYTE *)ycp + y * line_size );

        my_convert( ycp2, pixelp2, w, pixel_line_size, convert_mode );
    }
}

//---------------------------------------------------------------------
// [Xbh] YC48 -> YC48iójiFϊꍇj
//
// ̊֐ƂقƂǓ悤ȓe
//
// convert_mode
// 0: BT.601 oBT.709
// 1: BT.709 oBT.601
//---------------------------------------------------------------------
void __attribute__((force_align_arg_pointer)) thread_yc2pixel_yc48_convert( int thread_id, int thread_num, void *param1, void *param2 )
{
    // l[
    const COLOR_PROC_INFO *cpip  = (COLOR_PROC_INFO *)param1;
    const int convert_mode = *(int *)param2 == 0 ? 1 : 0;

    // [JϐɃRs[
    const int h         = cpip->h;
    const int w         = cpip->w;
    const int line_size = cpip->line_size;

    // ̂ŁAXbh𔼕ɂė̏Ƀ\[X
    if( thread_id < 0 || thread_num <= 0 ) return; // 肦ȂlȂ̂
    if( thread_id & 0x1 ) return; // eîݎcj
    thread_id = thread_id >> 1;
    thread_num = (thread_num + 1 ) >> 1;

    // Xbhc̃TCY葽璲
    if( thread_num > h ) thread_num = h;
    if( thread_id >= thread_num ) return;

    // Xbh̉摜郉CvZ
    const int y_start = (h *  thread_id     ) / thread_num;
    const int y_end   = (h * (thread_id + 1)) / thread_num;

    // ϐvZ
    const int pixel_line_size = (w * sizeof(PIXEL_YC) + 3) & (~3L); // 4̔{ɐ؂グ
    PIXEL_YC *ycp = cpip->ycp;
    const unsigned char *pixelp  = (unsigned char *)cpip->pixelp;

    for( int y = y_start; y < y_end; y++ )
    {
        PIXEL_YC *pixelp2 = (PIXEL_YC *)( (unsigned char *)pixelp + y * pixel_line_size );
        const PIXEL_YC *ycp2 = (PIXEL_YC *)( (BYTE *)ycp + y * line_size );

        my_convert( pixelp2, ycp2, w, line_size, convert_mode );
    }
}

//---------------------------------------------------------------------
// Rs[̋ʕ
//---------------------------------------------------------------------
void __attribute__((force_align_arg_pointer)) my_copy( BYTE *dst, const BYTE *src, const int w, const int src_length  )
{
   // o͑̃|C^16byteACgɑB
    // 蔲̂߂xȏ̉KvB
    if( w > 64 )
    {
        _mm_prefetch( (BYTE *)src + src_length, _MM_HINT_NTA ); // ̃CvtFb`

        intptr_t padding = (-(intptr_t)dst) & 0xF;

        // 1oCg
        if( padding & 0x1 )
        {
            *dst++ = *src++;
        }
        // 2oCg
        if( padding & 0x2 )
        {
            *(WORD *)dst = *(WORD *)src;
            dst += 2;
            src += 2;
        }
        // 4oCg
        if( padding & 0x4 )
        {
            _mm_stream_si32( (int *)dst, *(int *)src );
            dst += 4;
            src += 4;
        }
        // 8oCg
        if( padding & 0x8 )
        {
            _mm_stream_si32( (int *)dst    , *(int *)src     );
            _mm_stream_si32( (int *)dst + 1, *(int *)src + 1 );
            dst += 8;
            src += 8;
        }

        // ACĝ32oCg
        int x = padding;
        for( ; x < w - 31; x += 32 )
        {
            _mm_prefetch( (BYTE *)src + src_length, _MM_HINT_NTA ); // ̃CvtFb`
            __m128i t0 = _mm_loadu_si128( (__m128i *)src     );
            __m128i t1 = _mm_loadu_si128( (__m128i *)src + 1 );
            _mm_stream_si128( (__m128i *)dst    , t0 );
            _mm_stream_si128( (__m128i *)dst + 1, t1 );
            dst += 32;
            src += 32;
        }

        padding = w - x;

        // [16oCg
        if( padding & 0x10 )
        {
            __m128i t = _mm_loadu_si128( (__m128i *)src );
            _mm_stream_si128( (__m128i *)dst, t );
            dst += 16;
            src += 16;
        }
        // [8oCg
        if( padding & 0x8 )
        {
            _mm_stream_si32( (int *)dst    , *(int *)src     );
            _mm_stream_si32( (int *)dst + 1, *(int *)src + 1 );
            dst += 8;
            src += 8;
        }
        // [4oCg
        if( padding & 0x4 )
        {
            _mm_stream_si32( (int *)dst, *(int *)src );
            dst += 4;
            src += 4;
        }
        // [2oCg
        if( padding & 0x2 )
        {
            *(WORD *)dst = *(WORD *)src;
            dst += 2;
            src += 2;
        }
        // [1oCg
        if( padding & 0x1 )
        {
            *dst++ = *src++;
        }

    }
    else
    {
        // ACgĂȂ
        // 1oCgPʂŃRs[
        for( int x = 0; x < w; x++ ) *dst++ = *src++;
    }
}

//---------------------------------------------------------------------
// [Xbh] YC48() -> YC48iRs[j
//---------------------------------------------------------------------
void __attribute__((force_align_arg_pointer)) thread_pixel2yc_yc48_copy( int thread_id, int thread_num, void *param1, void *param2 )
{
    // l[
    const COLOR_PROC_INFO *cpip  = (COLOR_PROC_INFO *)param1;

    // [JϐɃRs[
    const int h         = cpip->h;
    const int w         = cpip->w * sizeof(PIXEL_YC);
    const int line_size = cpip->line_size;

    // ̂ŁAXbh𔼕ɂė̏Ƀ\[X
    if( thread_id < 0 || thread_num <= 0 ) return; // 肦ȂlȂ̂
    if( thread_id & 0x1 ) return; // eîݎcj
    thread_id = thread_id >> 1;
    thread_num = (thread_num + 1 ) >> 1;

    // Xbhc̃TCY葽璲
    if( thread_num > h ) thread_num = h;
    if( thread_id >= thread_num ) return;

    // Xbh̉摜郉CvZ
    const int y_start = (h *  thread_id     ) / thread_num;
    const int y_end   = (h * (thread_id + 1)) / thread_num;

    // ϐvZ
    const int pixel_line_size = (w + 3) & (~3L); // 4̔{ɐ؂グ
    BYTE *ycp = (BYTE *)cpip->ycp;
    const unsigned char *pixelp  = (unsigned char *)cpip->pixelp;

    for( int y = y_start; y < y_end; y++ )
    {
        const BYTE *pixelp2 = (BYTE *)( (BYTE *)pixelp + y * pixel_line_size );
        BYTE *ycp2 = (BYTE *)( (BYTE *)ycp + y * line_size );

        my_copy( ycp2, pixelp2, w, pixel_line_size );
    }
}

//---------------------------------------------------------------------
// [Xbh] YC48 -> YC48(o) iRs[j
//---------------------------------------------------------------------
void __attribute__((force_align_arg_pointer)) thread_yc2pixel_yc48_copy( int thread_id, int thread_num, void *param1, void *param2 )
{
    // l[
    const COLOR_PROC_INFO *cpip  = (COLOR_PROC_INFO *)param1;

    // [JϐɃRs[
    const int h         = cpip->h;
    const int w         = cpip->w * sizeof(PIXEL_YC);
    const int line_size = cpip->line_size;

    // ̂ŁAXbh𔼕ɂė̏Ƀ\[X
    if( thread_id < 0 || thread_num <= 0 ) return; // 肦ȂlȂ̂
    if( thread_id & 0x1 ) return; // eîݎcj
    thread_id = thread_id >> 1;
    thread_num = (thread_num + 1 ) >> 1;

    // Xbhc̃TCY葽璲
    if( thread_num > h ) thread_num = h;
    if( thread_id >= thread_num ) return;

    // Xbh̉摜郉CvZ
    const int y_start = (h *  thread_id     ) / thread_num;
    const int y_end   = (h * (thread_id + 1)) / thread_num;

    // ϐvZ
    const int pixel_line_size = (w + 3) & (~3L); // 4̔{ɐ؂グ
    BYTE *ycp = (BYTE *)cpip->ycp;
    const unsigned char *pixelp  = (unsigned char *)cpip->pixelp;

    for( int y = y_start; y < y_end; y++ )
    {
        BYTE *pixelp2 = (BYTE *)( (unsigned char *)pixelp + y * pixel_line_size );
        const BYTE *ycp2 = (BYTE *)( (BYTE *)ycp + y * line_size );

        my_copy( pixelp2, ycp2, w, line_size );
    }
}

